package com.chelvc.framework.email;

import java.util.Map;
import java.util.Objects;

import ch.qos.logback.classic.Level;
import ch.qos.logback.classic.filter.ThresholdFilter;
import ch.qos.logback.classic.spi.ILoggingEvent;
import ch.qos.logback.classic.spi.IThrowableProxy;
import ch.qos.logback.core.AppenderBase;
import ch.qos.logback.core.Layout;
import ch.qos.logback.core.spi.FilterReply;
import com.chelvc.framework.base.util.SpringUtils;
import com.chelvc.framework.common.util.StringUtils;
import com.google.common.collect.Maps;
import lombok.Getter;
import lombok.Setter;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.stereotype.Component;

/**
 * Logback日志邮件发送器
 *
 * @author Woody
 * @date 2024/5/20
 */
@Getter
@Setter
@Component
public class LogbackEmailAppender extends AppenderBase<ILoggingEvent> implements ApplicationContextAware {
    /**
     * 邮件处理器实例
     */
    private static EmailHandler EMAIL_HANDLER;

    private String to;
    private String subject;
    private long interval = 60000;
    private boolean asynchronous = true;
    private Layout<ILoggingEvent> layout;
    private final Map<Integer, Long> hashes = Maps.newConcurrentMap();

    /**
     * 判断日志邮件发送器是否已准备就绪
     *
     * @return true/false
     */
    protected boolean isReady() {
        return this.layout != null && EMAIL_HANDLER != null && StringUtils.notEmpty(this.to);
    }

    /**
     * 获取邮件处理器
     *
     * @return 邮件处理器
     */
    protected EmailHandler getEmailHandler() {
        return Objects.requireNonNull(EMAIL_HANDLER);
    }

    /**
     * 获取事件hash值
     *
     * @param event 事件对象实例
     * @return hash值
     */
    protected int hash(ILoggingEvent event) {
        IThrowableProxy throwable = event.getThrowableProxy();
        int hash = throwable == null ? 0 : Objects.hash((Object[]) throwable.getStackTraceElementProxyArray());
        return 31 * Math.max(hash, 1) + Objects.hashCode(event.getLoggerName());
    }

    /**
     * 判断是否是重复事件
     *
     * @param event 事件对象实例
     * @return true/false
     */
    protected boolean isDuplicated(ILoggingEvent event) {
        int hash = this.hash(event);
        long now = System.currentTimeMillis();
        Long timestamp = this.hashes.put(hash, now);
        return timestamp != null && now - timestamp <= this.interval;
    }

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        EMAIL_HANDLER = applicationContext.getBean(EmailHandler.class);
    }

    @Override
    protected void append(ILoggingEvent event) {
        if (!this.isReady() || (event.getLevel() == Level.ERROR && this.isDuplicated(event))) {
            return;
        }

        // 构建邮件体
        String type = this.layout.getContentType();
        String content = this.layout.doLayout(event);
        Body body = Body.builder().to(this.to).type(type).subject(this.subject).content(content).build();

        // 异步发送邮件
        EmailHandler handler = this.getEmailHandler();
        if (this.asynchronous) {
            this.context.getScheduledExecutorService().execute(() -> handler.send(body));
        } else {
            handler.send(body);
        }
    }

    /**
     * 日志邮件过滤器实现
     */
    public static class Filter extends ThresholdFilter {
        private String[] includes;
        private String[] excludes;

        /**
         * 判断事件是否满足邮件发送条件
         *
         * @param event 事件对象实例
         * @return true/false
         */
        protected boolean matches(ILoggingEvent event) {
            String logger = event.getLoggerName();
            return (StringUtils.isEmpty(this.excludes) || !SpringUtils.isPath(logger, this.excludes))
                    && (StringUtils.isEmpty(this.includes) || SpringUtils.isPath(logger, this.includes));
        }

        /**
         * 设置包含的日志包路径
         *
         * @param includes 包路径
         */
        public void setIncludes(String includes) {
            this.includes = StringUtils.ifEmpty(includes, StringUtils::splitActives);
        }

        /**
         * 设置排除的日志包路径
         *
         * @param excludes 包路径
         */
        public void setExcludes(String excludes) {
            this.excludes = StringUtils.ifEmpty(excludes, StringUtils::splitActives);
        }

        @Override
        public FilterReply decide(ILoggingEvent event) {
            if (!this.isStarted()) {
                return FilterReply.NEUTRAL;
            }

            FilterReply reply = super.decide(event);
            if (reply == FilterReply.NEUTRAL && this.matches(event)) {
                return FilterReply.NEUTRAL;
            }
            return FilterReply.DENY;
        }
    }
}
